import asyncio
import os
from functools import wraps
from typing import Tuple

from surf_tls.utils.asset import generate_asset_name
from surf_tls.utils.asset import root_dir
from surf_tls.exceptions.exceptions import SurfTLSClientException
import httpx

# GitHub repository for surf-tls-client releases
owner = 'ClaritySolutionsLLC'
repo = 'surf-tls-client'
url = f'https://api.github.com/repos/{owner}/{repo}/releases/latest'
root_directory = root_dir()
GITHUB_TOKEN = os.getenv("GH_TOKEN")


def auto_retry(retries: int):
    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            attempt = 0
            while attempt <= retries:
                try:
                    return await func(*args, **kwargs)
                except Exception as e:
                    attempt += 1
                    if attempt > retries:
                        print(f">> Failed after {attempt} attempts with error: {e}")
                        raise e
                    await asyncio.sleep(0.1)

        return wrapper

    return decorator


@auto_retry(retries=3)
async def get_latest_release() -> Tuple[str, list]:
    """
    Fetches the latest release from the GitHub API.

    :return: Latest release tag name, and a list of assets
    """
    # Make a GET request to the GitHub API
    async with httpx.AsyncClient() as client:
        headers = {
            'Accept': 'application/vnd.github.v3+json',
            'User-Agent': 'surf-tls'
        }
        response = await client.get(url, headers=headers)

    # Check if the request was successful
    if response.status_code == 200:
        data = response.json()  # Parse the JSON data from the response
        version_num = data['tag_name'].replace('v', '')  # Return the tag name without the 'v' prefix
        if 'assets' not in data:
            raise SurfTLSClientException(f"Version {version_num} does not have any assets.")

        # Get assets
        assets = data['assets']
        return version_num, assets
    else:
        raise SurfTLSClientException(f"Failed to fetch the latest release. Status code: {response.status_code}")


async def download_and_save_asset(
        asset_url: str,
        asset_name: str,
        version: str
) -> None:
    # Download
    async with httpx.AsyncClient(follow_redirects=True) as client:
        headers = {
            'Accept': 'application/octet-stream',
            'User-Agent': 'surf-tls',
            'Connection': 'keep-alive'
        }
        if GITHUB_TOKEN:
            headers["Authorization"] = f"token {GITHUB_TOKEN}"
            print(">> Using GitHub token for authentication.")

        response = await client.get(asset_url, headers=headers)
        if response.status_code != 200:
            raise SurfTLSClientException(f"Failed to download asset {asset_name}. Status code: {response.status_code}")

        # Ensure dependencies directory exists
        dependencies_dir = os.path.join(root_directory, 'dependencies')
        os.makedirs(dependencies_dir, exist_ok=True)

        with open(os.path.join(dependencies_dir, asset_name), 'wb') as f:
            f.write(response.content)

        # Save version info
        await save_version_info(asset_name, version)


async def save_version_info(asset_name: str, version: str):
    """
    Save version info to a hidden .version file in root_dir/dependencies
    """
    dependencies_dir = os.path.join(root_directory, 'dependencies')
    os.makedirs(dependencies_dir, exist_ok=True)
    with open(os.path.join(dependencies_dir, '.version'), 'w') as f:
        f.write(f"{asset_name} {version}")


def delete_version_info():
    """
    Delete everything inside dependencies/.version
    """
    try:
        dependencies_dir = os.path.join(root_directory, 'dependencies')
        # Delete all files in dependencies
        for file in os.listdir(dependencies_dir):
            file_path = os.path.join(dependencies_dir, file)
            if os.path.isfile(file_path):
                os.remove(file_path)
    except FileNotFoundError:
        pass


def read_version_info():
    """
    Read version info from a hidden .version file in root_dir/dependencies
    """
    try:
        dependencies_dir = os.path.join(root_directory, 'dependencies')
        version_file = os.path.join(dependencies_dir, '.version')
        with open(version_file, 'r') as f:
            data = f.read()
            data = data.split(' ')
            return data[0], data[1]
    except FileNotFoundError:
        return None, None


async def download_if_necessary():
    version_num, asset_url = await get_latest_release()
    if not asset_url or not version_num:
        raise SurfTLSClientException(f"Version {version_num} does not have any assets.")

    asset_name = generate_asset_name(custom_part='surf-tls-client', version=version_num)
    dependencies_dir = os.path.join(root_directory, 'dependencies')
    
    # Check if asset name is in the list of assets in root dir/dependencies
    if os.path.exists(os.path.join(dependencies_dir, asset_name)):
        return

    download_url = [asset['browser_download_url'] for asset in asset_url if asset['name'] == asset_name]
    if len(download_url) == 0:
        raise SurfTLSClientException(f"Unable to find asset {asset_name} for version {version_num}.")

    download_url = download_url[0]
    await download_and_save_asset(download_url, asset_name, version_num)


async def update_if_necessary():
    current_asset, current_version = read_version_info()
    if not current_asset or not current_version:
        raise SurfTLSClientException("Unable to read version info, no TLS libs found, use download_if_necessary()")

    version_num, asset_url = await get_latest_release()
    if not asset_url or not version_num:
        raise SurfTLSClientException(f"Version {version_num} does not have any assets.")

    if version_num != current_version:
        print(f">> Current version {current_version} is outdated, downloading the latest TLS release...")
        await download_if_necessary()


if __name__ == "__main__":
    asyncio.run(update_if_necessary())



import asyncio
import os
import sys
import ctypes

from surf_tls.exceptions.exceptions import SurfTLSClientException
from surf_tls.utils.asset import generate_asset_name, root_dir
from surf_tls.updater.file_fetch import read_version_info


async def check_and_download_dependencies():
    """
    Check if the dependencies folder is empty and download necessary files if it is.
    """
    root_directory = root_dir()
    dependencies_dir = f'{root_directory}/dependencies'
    if not os.path.exists(dependencies_dir):
        os.makedirs(dependencies_dir, exist_ok=True)
    
    contains_anything = [file for file in os.listdir(dependencies_dir) if not file.startswith('.')]
    if len(contains_anything) == 0:
        print(">> Dependencies folder is empty. Please build the surf-tls-client shared library.")
        print(">> Run: cd cffi_dist && go build -buildmode=c-shared -o ../dependencies/surf-tls-client.so")


def run_async_task(task):
    """
    Run an asynchronous task taking into account the current event loop.
    :param task: Coroutine to run.
    """
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(task)
    else:
        if loop.is_running():
            asyncio.ensure_future(task)
        else:
            loop.run_until_complete(task)


def load_asset():
    """
    Load the asset and return its name.
    :return: Name of the asset.
    """
    # Check if dependencies folder exists
    # Dependencies are in surf_tls/dependencies/ relative to the package
    package_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    dependencies_dir = os.path.join(package_dir, 'dependencies')
    
    if not os.path.exists(dependencies_dir):
        os.makedirs(dependencies_dir, exist_ok=True)

    # Try to read version info first
    current_asset, current_version = read_version_info()
    
    if current_asset and current_version:
        # Use the asset name from version info
        asset_name = current_asset
    else:
        # Fallback to generating asset name (for local development)
        asset_name = generate_asset_name()
    
    asset_path = os.path.join(dependencies_dir, asset_name)
    if not os.path.exists(asset_path):
        raise SurfTLSClientException(
            f"Unable to find asset {asset_name} at {asset_path}. "
            f"Please run: await surf_tls.download_if_necessary() or build locally with: ./build.sh"
        )

    return asset_name


def initialize_library():
    """
    Initialize and return the library.
    :return: Loaded library object.
    """
    try:
        asset_name = load_asset()
        # Dependencies are in surf_tls/dependencies/ relative to the package
        package_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        dependencies_dir = os.path.join(package_dir, 'dependencies')
        asset_path = os.path.join(dependencies_dir, asset_name)
        library = ctypes.cdll.LoadLibrary(asset_path)
        return library
    except SurfTLSClientException as e:
        print(f">> Failed to load the Surf TLS Client asset: {e}")
        return None
    except OSError as e:
        print(f">> Failed to load the library: {e}")
        if sys.platform == "darwin":
            print(">> If you're on macOS, you may need to allow the library to be loaded:")
            print(">> System Preferences > Security & Privacy > General > Allow")
        from surf_tls.utils.asset import generate_asset_name
        package_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        dependencies_dir = os.path.join(package_dir, 'dependencies')
        asset_path = os.path.join(dependencies_dir, generate_asset_name())
        print(f">> Asset path attempted: {asset_path}")
        return None


library = initialize_library()

if library is None:
    # Create dummy functions that will raise an error when called
    def _error_function(*args, **kwargs):
        raise SurfTLSClientException("Surf TLS Client library not loaded. Please build the shared library first by running: ./build.sh")
    
    request = _error_function
    free_memory = _error_function
else:
    # Define the request function from the shared package
    request = library.request
    request.argtypes = [ctypes.c_char_p]
    request.restype = ctypes.c_char_p

    free_memory = library.freeMemory
    free_memory.argtypes = [ctypes.c_char_p]
    free_memory.restype = ctypes.c_char_p



from .sessions import Session
from .__version__ import __version__
from .updater.file_fetch import (
    download_if_necessary,
    update_if_necessary,
    get_latest_release,
    read_version_info
)

__all__ = [
    'Session', 
    '__version__',
    'download_if_necessary',
    'update_if_necessary',
    'get_latest_release',
    'read_version_info'
]



import ctypes
import os
import platform
import sys
import distro


def root_dir():
    """Returns the root directory of the surf_tls package."""
    # Get the absolute path of the current file
    current_file_path = os.path.abspath(__file__)
    # Get the directory of the current file
    current_dir_path = os.path.dirname(current_file_path)
    # Navigate up one level to get the 'root_dir'
    root_dir = os.path.dirname(current_dir_path)
    return root_dir


def get_system_platform():
    return sys.platform


def get_distro():
    try:
        return distro.id()
    except:
        return "linux"


def generate_asset_name(
        custom_part: str = 'surf-tls-client',
        version: str = '0.1.0'
) -> str:
    """
    Generates an asset name based on the current platform and architecture.

    :param custom_part: Custom part of the name specified by the user, e.g., 'surf-tls-client'
    :param version: Version number specified by the user, e.g., '0.1.0'
    :return: Formatted asset name string
    """
    # Get the system's OS name and architecture
    system_os = platform.system().lower()
    architecture = platform.machine().lower()
    sys_platform = get_system_platform()

    # Correct the platform checks and architecture determination
    if sys_platform == 'darwin':
        file_extension = '.dylib'
        asset_arch = 'arm64' if architecture == "arm64" else 'amd64'
    elif sys_platform in ('win32', 'cygwin'):
        file_extension = '.dll'
        asset_arch = '64' if 8 == ctypes.sizeof(ctypes.c_voidp) else '32'
    else:
        # Linux
        file_extension = '.so'

        if architecture == "aarch64":
            asset_arch = 'arm64'
        elif "x86" in architecture:
            asset_arch = 'amd64'
        else:
            asset_arch = 'amd64'

        if system_os == 'linux':
            distro_name = get_distro()
            if distro_name.lower() in {"ubuntu", "debian"}:
                system_os = f"{system_os}-ubuntu"

    return f"{custom_part}-{system_os}-{asset_arch}-{version}{file_extension}"


if __name__ == "__main__":
    # Example usage:
    custom_part = 'surf-tls-client'
    version = '0.1.0'
    asset_name = generate_asset_name(custom_part, version)
    print(f">> Asset name: {asset_name}")



class CaseInsensitiveDict(dict):
    """A case-insensitive dictionary."""
    
    def __init__(self, data=None, **kwargs):
        super().__init__()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __setitem__(self, key, value):
        super().__setitem__(key.lower() if isinstance(key, str) else key, value)

    def __getitem__(self, key):
        return super().__getitem__(key.lower() if isinstance(key, str) else key)

    def __contains__(self, key):
        return super().__contains__(key.lower() if isinstance(key, str) else key)

    def get(self, key, default=None):
        return super().get(key.lower() if isinstance(key, str) else key, default)

    def update(self, data=None, **kwargs):
        if data:
            if isinstance(data, dict):
                for key, value in data.items():
                    self[key] = value
            else:
                for key, value in data:
                    self[key] = value
        for key, value in kwargs.items():
            self[key] = value

    def pop(self, key, default=None):
        return super().pop(key.lower() if isinstance(key, str) else key, default)



from .file_fetch import (
    download_if_necessary,
    update_if_necessary,
    get_latest_release,
    read_version_info
)

__all__ = ['download_if_necessary', 'update_if_necessary', 'get_latest_release', 'read_version_info']



from typing import Union, Dict, Any
import json
from .cookies import cookiejar_from_dict
from .utils.structures import CaseInsensitiveDict
from typing import Optional
from requests.exceptions import HTTPError


class Response:
    """Represents the response to an HTTP request."""

    def __init__(self):
        self.url: Optional[str] = None
        self.status_code: Optional[int] = None  # The HTTP status code.
        self.text: Optional[str] = None  # The text content of the response.
        self.headers: CaseInsensitiveDict = CaseInsensitiveDict()  # Case-insensitive response headers.
        self.cookies = cookiejar_from_dict({})  # Cookies sent back by the server.
        self._content: Optional[bytes] = None  # The byte content of the response.
        self._content_consumed: bool = False  # Tracks if the content has been consumed.
        self.history = []
        self.used_protocol: Optional[str] = None  # HTTP protocol version (HTTP/1.1, HTTP/2, HTTP/3)

    def __enter__(self):
        return self

    def __repr__(self):
        return f"<Response [{self.status_code}]>"

    def json(self, **kwargs) -> Union[Dict, list]:
        """Parses the text content of the response to JSON."""
        return json.loads(self.text, **kwargs)

    def raise_for_status(self):
        """Raises an HTTPError if the HTTP request returned an unsuccessful status code."""
        if 400 <= self.status_code < 500:
            raise HTTPError(f'Client Error: {self.status_code} for url: {self.url}')
        elif 500 <= self.status_code < 600:
            raise HTTPError(f'Server Error: {self.status_code} for url: {self.url}')

    @property
    def content(self) -> bytes:
        """Lazily loads the content of the response, in bytes."""
        if self._content is None:
            if self._content_consumed:
                raise RuntimeError("The content for this response was already consumed.")
            self._content = self.text.encode() if self.status_code != 0 else b""
            self._content_consumed = True
        return self._content


def build_response(res: Dict[str, Any], res_cookies) -> Response:
    """Builds and returns a Response object from given data."""
    response = Response()
    response.url = res.get("target")  # Extract the target URL from the response data.
    response.status_code = res.get("status", 0)  # Default to 0 if status is not provided.
    response.text = res.get("body", "")  # Default to empty string if body is not provided.
    response.used_protocol = res.get("usedProtocol", "unknown")  # HTTP protocol version

    # Process headers, ensuring single values are not wrapped in a list.
    response_headers = CaseInsensitiveDict()
    for key, value in res.get("headers", {}).items():
        response_headers[key] = value[0] if isinstance(value, list) and len(value) == 1 else value
    response.headers = response_headers

    response.cookies = res_cookies  # Assign the provided cookies to the response.
    return response



# Builtins
import asyncio
from typing import Any, Optional, Union
from json import dumps, loads
import urllib.parse
import base64
import ctypes

from .c.cffi import request, free_memory
from .cookies import cookiejar_from_dict, merge_cookies, extract_cookies_to_jar
from .exceptions.exceptions import SurfTLSClientException
from .utils.structures import CaseInsensitiveDict
from .__version__ import __version__
from .response import build_response
from .utils.session_utils import random_session_id


class Session:
    def __init__(
            self,
            ja3_string: Optional[str] = None,
            ja3_hello_id: Optional[str] = None,
            h2_settings: Optional[dict] = None,
            h2_settings_order: Optional[list] = None,
            pseudo_header_order: Optional[list] = None,
            header_order: Optional[list] = None,
            random_tls_extension_order: Optional[bool] = False,
            force_http1: Optional[bool] = False,
            force_http2: Optional[bool] = False,
            force_http3: Optional[bool] = False,
            disable_http3: Optional[bool] = False,
            insecure_skip_verify: Optional[bool] = False,
            debug: Optional[bool] = False,
            transport_options: Optional[dict] = None,
            connect_headers: Optional[dict] = None
    ) -> None:
        self._session_id = random_session_id()
        
        # Case-insensitive dictionary of headers, send on each request
        self.headers = CaseInsensitiveDict(
            {
                "User-Agent": f"surf-tls/{__version__}",
                "Accept-Encoding": "gzip, deflate, br",
                "Accept": "*/*",
                "Connection": "keep-alive",
            }
        )

        self.proxies = {}
        self.params = {}
        self.cookies = cookiejar_from_dict({})
        self.timeout_seconds = 30

        # TLS fingerprinting settings
        self.ja3_string = ja3_string
        self.ja3_hello_id = ja3_hello_id
        
        # HTTP/2 settings
        self.h2_settings = h2_settings
        self.h2_settings_order = h2_settings_order
        self.pseudo_header_order = pseudo_header_order
        self.header_order = header_order
        
        # Protocol settings
        self.random_tls_extension_order = random_tls_extension_order
        self.force_http1 = force_http1
        self.force_http2 = force_http2
        self.force_http3 = force_http3
        self.disable_http3 = disable_http3
        self.insecure_skip_verify = insecure_skip_verify
        
        # Other settings
        self.debug = debug
        self.transport_options = transport_options
        self.connect_headers = connect_headers

        self.loop = asyncio.get_event_loop()

    @property
    def timeout(self):
        return self.timeout_seconds

    @timeout.setter
    def timeout(self, seconds):
        self.timeout_seconds = seconds

    async def execute_request(
            self,
            method: str,
            url: str,
            params: Optional[dict] = None,
            data: Optional[Union[str, dict]] = None,
            headers: Optional[dict] = None,
            cookies: Optional[dict] = None,
            json: Optional[dict] = None,
            allow_redirects: Optional[bool] = True,
            timeout_seconds: Optional[int] = None,
            timeout: Optional[int] = None,
            proxy: Optional[dict] = None,
            is_byte_response: Optional[bool] = False
    ):
        # Timeout
        timeout_seconds = timeout or timeout_seconds or self.timeout_seconds
        del timeout  # deleting alias to stop further usage

        # History for redirects
        history = []

        # Prepare URL - add params to url
        if params is not None:
            url = f"{url}?{urllib.parse.urlencode(params, doseq=True)}"

        # Prepare request body
        if data is None and json is not None:
            if type(json) in [dict, list]:
                json = dumps(json)
            request_body = json
            content_type = "application/json"
        elif data is not None and type(data) not in [str, bytes]:
            request_body = urllib.parse.urlencode(data, doseq=True)
            content_type = "application/x-www-form-urlencoded"
        else:
            request_body = data
            content_type = None
        
        # Set content type if it isn't set
        if content_type is not None and "content-type" not in self.headers:
            self.headers["Content-Type"] = content_type

        # Headers
        if self.headers is None:
            headers = CaseInsensitiveDict(headers)
        elif headers is None:
            headers = self.headers
        else:
            merged_headers = CaseInsensitiveDict(self.headers)
            merged_headers.update(headers)
            none_keys = [k for (k, v) in merged_headers.items() if v is None or k is None]
            for key in none_keys:
                del merged_headers[key]
            headers = merged_headers

        # Cookies
        cookies = cookies or {}
        cookies = merge_cookies(self.cookies, cookies)
        request_cookies = [
            {'domain': c.domain, 'expires': c.expires, 'name': c.name, 'path': c.path,
             'value': c.value.replace('"', "")}
            for c in cookies
        ]

        # Proxy
        proxy = proxy or self.proxies
        if type(proxy) is dict and "http" in proxy:
            proxy = proxy["http"]
        elif type(proxy) is str:
            proxy = proxy
        else:
            proxy = ""

        while True:
            # Build request payload
            is_byte_request = isinstance(request_body, (bytes, bytearray))
            request_payload = {
                "sessionId": self._session_id,
                "followRedirects": allow_redirects,
                "forceHttp1": self.force_http1,
                "forceHttp2": self.force_http2,
                "forceHttp3": self.force_http3,
                "disableHttp3": self.disable_http3,
                "withDebug": self.debug,
                "headers": dict(headers),
                "headerOrder": self.header_order,
                "insecureSkipVerify": self.insecure_skip_verify,
                "isByteRequest": is_byte_request,
                "isByteResponse": is_byte_response,
                "proxyUrl": proxy,
                "requestUrl": url,
                "requestMethod": method,
                "requestBody": base64.b64encode(request_body).decode() if is_byte_request else (request_body or ""),
                "requestCookies": request_cookies,
                "timeoutSeconds": timeout_seconds,
                "transportOptions": self.transport_options,
                "connectHeaders": self.connect_headers,
                "ja3String": self.ja3_string or "",
                "ja3HelloID": self.ja3_hello_id or "",
                "h2Settings": self.h2_settings or {},
                "h2SettingsOrder": self.h2_settings_order or [],
                "pseudoHeaderOrder": self.pseudo_header_order or [],
                "withRandomTLSExtensionOrder": self.random_tls_extension_order
            }

            loop = asyncio.get_event_loop()
            # Call the C function
            response = await loop.run_in_executor(None, request, dumps(request_payload).encode('utf-8'))

            # Dereference the pointer to a byte array
            response_bytes = ctypes.string_at(response)
            # Convert byte array to string (surf-tls-client returns json)
            response_string = response_bytes.decode('utf-8')
            # Convert response string to json
            response_object = loads(response_string)
            # Free the memory
            await loop.run_in_executor(None, free_memory, response_object['id'].encode('utf-8'))

            # Error handling
            if response_object["status"] == 0:
                raise SurfTLSClientException(response_object["body"])
            
            # Set response cookies
            response_cookie_jar = extract_cookies_to_jar(
                request_url=url,
                request_headers=headers,
                cookie_jar=cookies,
                response_headers=response_object["headers"]
            )
            
            # Build response class
            current_response = build_response(response_object, response_cookie_jar)
            
            # Check for redirect
            if allow_redirects:
                if 'Location' in (resp_headers := current_response.headers) and current_response.status_code in (
                    300, 301, 302, 303, 307, 308
                ):
                    history.append(current_response)
                    url = resp_headers['Location']
                else:
                    break
            else:
                break

        # Assign the history to the final response
        current_response.history = history
        return current_response

    async def get(self, url: str, **kwargs: Any):
        """Sends a GET request"""
        return await self.execute_request(method="GET", url=url, **kwargs)

    async def options(self, url: str, **kwargs: Any):
        """Sends an OPTIONS request"""
        return await self.execute_request(method="OPTIONS", url=url, **kwargs)

    async def head(self, url: str, **kwargs: Any):
        """Sends a HEAD request"""
        return await self.execute_request(method="HEAD", url=url, **kwargs)

    async def post(self, url: str, data: Optional[Union[str, dict]] = None, json: Optional[dict] = None, **kwargs: Any):
        """Sends a POST request"""
        return await self.execute_request(method="POST", url=url, data=data, json=json, **kwargs)

    async def put(self, url: str, data: Optional[Union[str, dict]] = None, json: Optional[dict] = None, **kwargs: Any):
        """Sends a PUT request"""
        return await self.execute_request(method="PUT", url=url, data=data, json=json, **kwargs)

    async def patch(self, url: str, data: Optional[Union[str, dict]] = None, json: Optional[dict] = None, **kwargs: Any):
        """Sends a PATCH request"""
        return await self.execute_request(method="PATCH", url=url, data=data, json=json, **kwargs)

    async def delete(self, url: str, **kwargs: Any):
        """Sends a DELETE request"""
        return await self.execute_request(method="DELETE", url=url, **kwargs)
